// Copyright 2020 CoreOS, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

mod cmdline;
mod kargs;
mod rootmap;
mod stream_hash;

use anyhow::{Context, Result};

use crate::cmdline::*;

fn main() -> Result<()> {
    let config = cmdline::parse_args().context("parsing arguments")?;

    match config {
        Config::Kargs(c) => kargs::kargs(&c),
        Config::RootMap(c) => rootmap::rootmap(&c),
        Config::StreamHash(c) => stream_hash::stream_hash(&c),
    }
}
